from yowsup.layers.noise.workers.handshake import WANoiseProtocolHandshakeWorker
from yowsup.layers import YowLayer, EventCallback
from yowsup.layers.auth.layer_authentication import YowAuthenticationProtocolLayer
from yowsup.layers.network.layer import YowNetworkLayer
from yowsup.layers.noise.layer_noise_segments import YowNoiseSegmentsLayer
from yowsup.config.manager import ConfigManager
from yowsup.env.env import YowsupEnv

from consonance.protocol import WANoiseProtocol
from consonance.config.client import ClientConfig
from consonance.config.useragent import UserAgentConfig
from consonance.streams.segmented.blockingqueue import BlockingQueueSegmentedStream
from consonance.structs.keypair import KeyPair

import threading
import logging


logger = logging.getLogger(__name__)
try:
    import Queue
except ImportError:
    import queue as Queue


class YowNoiseLayer(YowLayer):
    DEFAULT_PUSHNAME = "yowsup"
    HEADER = b'WA\x02\x01'
    EDGE_HEADER = b'ED\x00\x01'

    def __init__(self):
        super(YowNoiseLayer, self).__init__()
        self._wa_noiseprotocol = WANoiseProtocol(
            2, 1, protocol_state_callbacks=self._on_protocol_state_changed
        )  # type: WANoiseProtocol

        self._handshake_worker = None
        self._stream = BlockingQueueSegmentedStream()  # type: BlockingQueueSegmentedStream
        self._read_buffer = bytearray()
        self._flush_lock = threading.Lock()
        self._incoming_segments_queue = Queue.Queue()
        self._config_manager = ConfigManager()
        self._username = None
        self._rs = None

    def __str__(self):
        return "Noise Layer"

    @EventCallback(YowNetworkLayer.EVENT_STATE_DISCONNECTED)
    def on_disconnected(self, event):
        self._wa_noiseprotocol.reset()

    @EventCallback(YowAuthenticationProtocolLayer.EVENT_AUTH)
    def on_auth(self, event):
        logger.debug("Received auth event")
        credentials = event.getArg("credentials")
        self._username, local_static = int(credentials[0]), credentials[1]
        config = self._config_manager.load(self._username)  # type: yowsup.config.v1.config.Config
        # event's keypair will override config's keypair
        local_static = local_static or config.client_static_keypair
        if type(local_static) is bytes:
            local_static = KeyPair.from_bytes(local_static)
        assert type(local_static) is KeyPair
        passive = event.getArg('passive')

        self.setProp(YowNoiseSegmentsLayer.PROP_ENABLED, False)

        if config.edge_routing_info:
            self.toLower(self.EDGE_HEADER)
            self.setProp(YowNoiseSegmentsLayer.PROP_ENABLED, True)
            self.toLower(config.edge_routing_info)
            self.setProp(YowNoiseSegmentsLayer.PROP_ENABLED, False)

        self.toLower(self.HEADER)
        self.setProp(YowNoiseSegmentsLayer.PROP_ENABLED, True)

        remote_static = config.server_static_public

        self._rs = remote_static
        yowsupenv = YowsupEnv.getCurrent()
        client_config = ClientConfig(
            username=self._username,
            passive=passive,
            useragent=UserAgentConfig(
                platform=0,
                app_version=yowsupenv.getVersion(),
                mcc=config.mcc or "000",
                mnc=config.mnc or "000",
                os_version=yowsupenv.getOSVersion(),
                manufacturer=yowsupenv.getManufacturer(),
                device=yowsupenv.getDeviceName(),
                os_build_number=yowsupenv.getOSVersion(),
                phone_id=config.fdid or "",
                locale_lang="en",
                locale_country="US"
            ),
            pushname=config.pushname or self.DEFAULT_PUSHNAME,
            short_connect=True
        )
        if not self._in_handshake():
            logger.debug("Performing handshake [username= %d, passive=%s]" % (self._username, passive) )
            self._handshake_worker = WANoiseProtocolHandshakeWorker(
                self._wa_noiseprotocol, self._stream, client_config, local_static, remote_static,
            )
            logger.debug("Starting handshake worker")
            self._stream.set_events_callback(self._handle_stream_event)
            self._handshake_worker.start()

    def _in_handshake(self):
        """
        :return:
        :rtype: bool
        """
        return self._wa_noiseprotocol.state == WANoiseProtocol.STATE_HANDSHAKE

    def _on_protocol_state_changed(self, state):
        if state == WANoiseProtocol.STATE_TRANSPORT:
            if self._rs != self._wa_noiseprotocol.rs:
                config = self._config_manager.load(self._username)
                config.server_static_public = self._wa_noiseprotocol.rs
                self._config_manager.save(config)
                self._rs = self._wa_noiseprotocol.rs
            self._flush_incoming_buffer()

    def _handle_stream_event(self, event):
        if event == BlockingQueueSegmentedStream.EVENT_WRITE:
            self.toLower(self._stream.get_write_segment())
        elif event == BlockingQueueSegmentedStream.EVENT_READ:
            self._stream.put_read_segment(self._incoming_segments_queue.get(block=True))

    def send(self, data):
        """
        :param data:
        :type data: bytearray | bytes
        :return:
        :rtype:
        """
        data = bytes(data) if type(data) is not bytes else data
        self._wa_noiseprotocol.send(data)

    def _flush_incoming_buffer(self):
        self._flush_lock.acquire()
        while self._incoming_segments_queue.qsize():
            self.toUpper(self._wa_noiseprotocol.receive())
        self._flush_lock.release()

    def receive(self, data):
        """
        :param data:
        :type data: bytes
        :return:
        :rtype:
        """
        self._incoming_segments_queue.put(data)
        if not self._in_handshake():
            self._flush_incoming_buffer()



from yowsup.layers.protocol_messages.protocolentities.protomessage import ProtomessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_media import MediaAttributes
from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes

import logging
logger = logging.getLogger(__name__)


class MediaMessageProtocolEntity(ProtomessageProtocolEntity):
    TYPE_MEDIA_IMAGE = "image"
    TYPE_MEDIA_VIDEO = "video"
    TYPE_MEDIA_AUDIO = "audio"
    TYPE_MEDIA_CONTACT = "contact"
    TYPE_MEDIA_LOCATION = "location"
    TYPE_MEDIA_DOCUMENT = "document"
    TYPE_MEDIA_GIF = "gif"
    TYPE_MEDIA_PTT = "ptt"
    TYPE_MEDIA_URL = "url"

    TYPES_MEDIA = (
        TYPE_MEDIA_IMAGE, TYPE_MEDIA_AUDIO, TYPE_MEDIA_VIDEO,
        TYPE_MEDIA_CONTACT, TYPE_MEDIA_LOCATION, TYPE_MEDIA_DOCUMENT,
        TYPE_MEDIA_GIF, TYPE_MEDIA_PTT, TYPE_MEDIA_URL
    )

    def __init__(self, media_type, media_message_attrs, message_attrs):
        """
        :type media_type: str
        :type media_message_attrs: MediaAttributes
        :type message_attrs: MessageAttributes
        """
        super(MediaMessageProtocolEntity, self).__init__("media", message_attrs)
        self.media_type = media_type  # type: str

        if media_message_attrs.context_info:
            self.context_stanza_id = media_message_attrs.context_info.stanza_id
            self.context_participant = media_message_attrs.context_info.participant
            self.context_quoted_message = media_message_attrs.context_info.quoted_message
            self.context_edit_version = media_message_attrs.context_info.edit_version
            self.context_remote_jid = media_message_attrs.context_info.remote_jid
            self.context_mentioned_jid = media_message_attrs.context_info.mentioned_jid
            self.context_revoke_message = media_message_attrs.context_info.revoke_message

    def __str__(self):
        out = super(MediaMessageProtocolEntity, self).__str__()
        out += "\nmediatype=%s" % self.media_type
        return out

    @property
    def context_stanza_id(self):
        return self.proto.stanza_id

    @context_stanza_id.setter
    def context_stanza_id(self, value):
        self.proto.context_info.stanza_id = value

    @property
    def context_participant(self):
        return self.proto.context_info.participant

    @context_participant.setter
    def context_participant(self, value):
        self.proto.context_info = value

    @property
    def context_quoted_message(self):
        return self.proto.context_info.quoted_message

    @context_quoted_message.setter
    def context_quoted_message(self, value):
        self.proto.context_info.quoted_message = value

    @property
    def context_edit_version(self):
        return self.proto.context_info.edit_version

    @context_edit_version.setter
    def context_edit_version(self, value):
        self.proto.context_info.edit_version = value

    @property
    def context_remote_jid(self):
        return self.proto.context_info.remote_jid

    @context_remote_jid.setter
    def context_remote_jid(self, value):
        self.proto.context_info.remote_jid = value

    @property
    def context_mentioned_jid(self):
        return self.proto.context_info.mentioned_jid

    @context_mentioned_jid.setter
    def context_mentioned_jid(self, value):
        self.proto.context_info.mentioned_jid = value

    @property
    def context_revoke_message(self):
        return self.proto.context_info.revoke_message

    @context_revoke_message.setter
    def context_revoke_message(self, value):
        self.proto.context_info.revoke_message = value

    @property
    def media_type(self):
        return self._media_type

    @media_type.setter
    def media_type(self, value):
        if value not in MediaMessageProtocolEntity.TYPES_MEDIA:
            logger.warn("media type: '%s' is not supported" % value)
        self._media_type = value

    def toProtocolTreeNode(self):
        node = super(MediaMessageProtocolEntity, self).toProtocolTreeNode()
        protoNode = node.getChild("proto")
        protoNode["mediatype"] = self.media_type
        return node

    @classmethod
    def fromProtocolTreeNode(cls, node):
        entity = ProtomessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = cls
        entity.media_type = node.getChild("proto")["mediatype"]
        return entity



class VideoAttributes(object):
    def __init__(self, width, height, seconds,
                 gif_playback=None, jpeg_thumbnail=None, gif_attribution=None, caption=None, streaming_sidecar=None):
        self._width = width
        self._height = height
        self._seconds = seconds
        self._gif_playback = gif_playback
        self._jpeg_thumbnail = jpeg_thumbnail
        self._gif_attribution = gif_attribution
        self._caption = caption
        self._streaming_sidecar = streaming_sidecar

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def seconds(self):
        return self._seconds

    @property
    def gif_playback(self):
        return self._gif_playback

    @property
    def jpeg_thumbnail(self):
        return self._jpeg_thumbnail

    @property
    def gif_attribution(self):
        return self._gif_attribution

    @property
    def caption(self):
        return self._caption

    @property
    def streaming_sidecar(self):
        return self._streaming_sidecar



from .iq_key_get import GetKeysIqProtocolEntity
from .iq_keys_set import SetKeysIqProtocolEntity
from .iq_keys_get_result import ResultGetKeysIqProtocolEntity
from .message_encrypted import EncryptedMessageProtocolEntity
from .enc import EncProtocolEntity
from .receipt_outgoing_retry import RetryOutgoingReceiptProtocolEntity
from .receipt_incoming_retry import RetryIncomingReceiptProtocolEntity
from .notification_encrypt_identitychange import IdentityChangeEncryptNotification
from .notification_encrypt_requestkeys import RequestKeysEncryptNotification
